

import os
from numpy import array

import soundfile as sf

from snippyng.util import Sound

folder_containing_current_file, _ = os.path.split(__file__)
data_folder = os.path.join(folder_containing_current_file, 'data')

DFLT_SR = 44100
DFLT_ALPHABET_STR = "#$%&()*+,-./0123456789:<=>?[\]^_abcdefghijklmnopqrstuvwxyz~"
DFLT_ALPHABET_WAV_FILE = os.path.join(data_folder, 'abcdef_etc.wav')


class TextToSound(object):
    def __init__(self, char_wf, sr=DFLT_SR):
        self.char_wf = char_wf
        self.sr = sr

    @classmethod
    def from_str_and_wav(cls, alphabet_str=DFLT_ALPHABET_STR, alphabet_wav_file=DFLT_ALPHABET_WAV_FILE,
                         wf_dtype='float64'):

        wf, sr = sf.read(alphabet_wav_file, dtype='float64')
        n_chars = len(alphabet_str)
        frm_per_char = len(wf) / n_chars

        char_wf = dict()
        for i, c in enumerate(alphabet_str):
            char_wf[c] = wf[int(i * frm_per_char): int((i + 1) * frm_per_char)]
        char_wf[' '] = array([0] * int(frm_per_char))  # add the space to the character

        return TextToSound(char_wf=char_wf, sr=sr)

    def text_to_wf(self, txt):
        assert set(txt + ' ').issubset(list(self.char_wf.keys())), \
            "Your text needs to only use space and the following characters:\n{}".format(''.join(char_wf))
        wf = []
        for c in txt:
            wf += list(self.char_wf[c])
        return array(wf)

    def text_to_sound(self, txt):
        return Sound(wf=self.text_to_wf(txt), sr=self.sr)



__author__ = 'thor'

import os
import pandas as pd
from pymongo import MongoClient
from pymongo.cursor import Cursor


from snippyng import util as sutil
from pynball.manip import reorder_columns_as
from snippyng.util import Sound
from typyst.trans import str_to_utf8_or_bust


class MgDacc(object):
    def __init__(self, db, collection, root_folder, path_field='_id', mg_client_kwargs={}):
        self.mgc = MongoClient(**mg_client_kwargs)[db][collection]
        self.root_folder = root_folder
        self.path_field = path_field

    def filepath_of(self, path):
        return str_to_utf8_or_bust(os.path.join(self.root_folder, path))

    def get_wf_and_sr(self, path, **kwargs):
        return sutil.wf_and_sr_from_filepath(self.filepath_of(path), **kwargs)

    def get_sound(self, path_or_doc, **kwargs):
        if not isinstance(path_or_doc, str):
            path_or_doc = path_or_doc.copy()
            file_path = path_or_doc.pop(self.path_field)
            kwargs = dict(kwargs, **path_or_doc)
            path_or_doc = file_path
        name = kwargs.pop('name', os.path.splitext(os.path.basename(path_or_doc))[0])
        try:
            wf, sr = self.get_wf_and_sr(path_or_doc, **kwargs)
        except TypeError:
            kwargs.pop('channels')
            kwargs.pop('frames')
            wf, sr = self.get_wf_and_sr(path_or_doc, **kwargs)
        return Sound(wf=wf, sr=sr, name=name)

    def get_sound_iterator(self, find_args={}, find_kwargs={}):
        """
        Util to flip through sounds.
        You can do, for example:
            sound_iterator = self.get_sound_iterator
        and then run the following several times:
            sound = sound_iterator.next(); sound.display_sound()
        """
        if not find_args and not find_kwargs:
            cursor = self.mgc.find()
        else:
            cursor = self.mgc.find(*find_args, **find_kwargs)
        return map(lambda x: self.get_sound(path_or_doc=x[self.path_field]), cursor)


class SegmentDacc(MgDacc):
    def __init__(self, db, collection, root_folder, path_field='_id', mg_client_kwargs={},
                 segment_field='segments', feat_field='fv', tag_field='tags', kv_tag_field='kv_tags'):
        super(SegmentDacc, self).__init__(db, collection, root_folder, path_field, mg_client_kwargs)
        self.segment_field = segment_field
        self.feat_field = feat_field
        self.tag_field = tag_field
        self.kv_tag_field = kv_tag_field

    def get_data_with_tags(self, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], Cursor):
            c = args[0]
        else:
            c = self.mgc.find(*args, **kwargs)
        d = list()
        for ci in c:
            for seg in ci['segments']:
                dd = {'path': ci[self.path_field], 'tags': ci[self.tag_field]}
                dd.update(seg['fv'])
                dd.update({'offset_s': seg['offset_s'], 'duration': seg['duration']})
                d += [dd]
        d = reorder_columns_as(pd.DataFrame(d), ['path', 'tags', 'offset_s', 'duration'])
        return d

    def get_data_with_kv_tags(self, *args, **kwargs):
        if 'kv_tag_keys' in list(kwargs.keys()):
            kv_tag_keys = kwargs.get('kv_tag_keys')
            kwargs.pop('kv_tag_keys')
        else:
            kv_tag_keys = ['move_direction', 'vehicle_type']

        if len(args) > 0 and isinstance(args[0], Cursor):
            c = args[0]
        else:
            c = self.mgc.find(*args, **kwargs)
        d = list()
        for ci in c:
            for seg in ci[self.segment_field]:
                dd = {'path': ci[self.path_field]}
                for tag_key in kv_tag_keys:
                    dd.update({tag_key: ci[self.kv_tag_field].get(tag_key, None)})
                dd.update(seg['fv'])
                dd.update({'offset_s': seg['offset_s'], 'duration': seg['duration']})
                d += [dd]
        d = reorder_columns_as(pd.DataFrame(d), ['path'] + kv_tag_keys + ['offset_s', 'duration'])
        return d

    # def get_sound(self, *args, **kwargs):
    #     # if len(args) > 0:
    #     #     kwargs['path_or_doc'] = args[0]
    #     return super(SegmentDacc, self).get_sound(path_or_doc=, **kwargs)

            # return super(SegmentDacc, self).get_sound(args[0], **kwargs)
        # return super(SegmentDacc, self).get_sound(path_or_doc=kwargs['path'],
        #                                           offset_s=kwargs['offset_s'],
        #                                           duration=kwargs['duration'])

    def get_segment_iterator(self, only_segments=True, fields=None, *args, **kwargs):
        cursor = self.mgc.find(*args, **kwargs)

        def segment_iterator():
            for d in cursor:
                segments = d.pop(self.segment_field)
                if segments is not None:
                    for dd in segments:
                        if not only_segments:
                            dd = dict(d, **dd)
                        if fields is None:
                            yield dd
                        else:
                            yield {k: v for k, v in dd.items() if k in fields}

        return segment_iterator()

    def get_sound_iterator(self, *args, **kwargs):
        """
        Util to flip through sounds.
        You can do, for example:
            sound_iterator = self.get_sound_iterator
        and then run the following several times:
            sound = sound_iterator.next(); sound.display_sound()
        """

        cursor = self.mgc.find(*args, **kwargs)
        return map(self.get_sound, cursor)



__author__ = 'thor'

from numpy import hstack, vstack, zeros, shape, mean, random, linspace
from numpy import argmax, argmin, flipud, percentile, ndarray, ceil
import numpy as np
import os
import re
import librosa
import librosa.display
import soundfile as sf
import wave
import contextlib

from IPython.display import Audio
import matplotlib.pyplot as plt
# import numpy as np
from scipy.signal import resample as scipy_signal_resample

# from functools import partial

from nippy.log import printProgress
from nippy.pfunc import filter_kwargs_to_func_arguments
import subprocess

default_sr = 44100

wav_text_info_exp = re.compile("^.*WAVEbextZ\x03\x00\x00([^\x00]+)")

TMP_FILE = 'ut_sound_util_tmp_file.wav'


def convert_to_wav(source_file, target_file=None, sample_rate=default_sr, print_stats=False):
    if target_file is None:
        folder, filename = os.path.split(source_file)
        extension_less_filename, ext = os.path.splitext(filename)
        target_file = os.path.join(folder, extension_less_filename + '.wav')
    if print_stats:
        return subprocess.call(['ffmpeg', '-i', source_file, '-ar', str(sample_rate), target_file])
    else:
        return subprocess.call(['ffmpeg', '-nostats', '-i', source_file, '-ar', str(sample_rate), target_file])


def complete_sref(sref):
    """
    Complete sref dict with missing fields, if any
    """
    sref = dict({'offset_s': 0.0}, **sref)  # takes care of making a copy (so doesn't overwrite sref)
    if 'duration' not in list(sref.keys()):
        if is_wav_file(sref['filepath']):
            sref['duration'] = get_duration_of_wav_file(sref['filepath'])
        else:
            sound = Sound.from_file(sref['filepath'])
            sref['duration'] = duration_of_wf_and_sr(sound.wf, sound.sr) - sref['offset_s']
    return sref


def sound_file_info_dict(filepath):
    filename = os.path.basename(filepath)
    (shortname, extension) = os.path.splitext(filename)
    d = {'filepath': filepath,
         'name': shortname,
         'size': os.path.getsize(filepath),
         'ext': extension[1:]
         }
    if extension == '.wav':
        with contextlib.closing(wave.open(filepath, 'r')) as f:
            d['channels'] = f.getnchannels()
            d['sample_width'] = f.getsampwidth()
            d['frames'] = f.getnframes()
            d['frame_rate'] = f.getframerate()
            d['duration'] = d['frames'] / float(d['frame_rate'])
        with open(filepath, 'r') as f:
            text_info = get_wav_text_info(f)
            if text_info is not None:
                d['inner_wav_text'] = text_info

    return d


def get_frame_rate_of_wav_file(filepath):
    with contextlib.closing(wave.open(filepath, 'r')) as f:
        return f.getframerate()


def get_duration_of_wav_file(filepath):
    with contextlib.closing(wave.open(filepath, 'r')) as f:
        return f.getnframes() / f.getframerate()


def is_mono(wf):
    return len(shape(wf)) == 1


def ensure_mono(wf):
    if is_mono(wf):
        return wf
    else:
        return mean(wf, axis=1)
        # return wf[:, 0]


def resample_wf(wf, sr, new_sr):
    # TODO: Replace using sox
    # tfm = sox.Transformer()
    # tfm.set_output_format(rate=44100)
    # tfm.build('Sample-01.wav', 'test.wav')

    # return round(len(wf) * new_sr / sr)
    return scipy_signal_resample(wf, num=int(round(len(wf) * new_sr / sr)))


def suffix_with_silence(wf, num_silence_pts):
    if is_mono(wf):
        return hstack([wf, zeros(num_silence_pts)])
    else:
        return vstack([wf, zeros((num_silence_pts, 2))])


def prefix_with_silence(wf, num_silence_pts):
    if is_mono(wf):
        return hstack([zeros(num_silence_pts), wf])
    else:
        return vstack([zeros((num_silence_pts, 2)), wf])


def stereo_to_mono_and_extreme_silence_cropping(source, target, subtype=None, print_progress=False):
    if os.path.isdir(source) and os.path.isdir(target):
        from glob import iglob
        if source[-1] != '/':
            source += '/'
        for i, filepath in enumerate(iglob(source + '*.wav')):
            filename = os.path.basename(filepath)
            if print_progress:
                printProgress("{}: {}".format(i, filename))
            stereo_to_mono_and_extreme_silence_cropping(
                filepath,
                os.path.join(target, filename)
            )
    else:
        wf, sr = wf_and_sr(source)
        wf = ensure_mono(wf)
        wf = crop_head_and_tail_silence(wf)
        sf.write(target, wf, samplerate=sr, subtype=subtype)


def get_wav_text_info(filespec):
    if isinstance(filespec, str):
        with open(filespec, 'r') as fd:
            m = wav_text_info_exp.match(fd.read())
            if m is not None:
                return m.groups()[0]
            else:
                return None
    else:
        m = wav_text_info_exp.match(filespec.read())
        if m is not None:
            return m.groups()[0]
        else:
            return None


def wf_and_sr(*args, **kwargs):
    if len(args) > 0:
        args_0 = args[0]
        if isinstance(args_0, str):
            kwargs['filepath'] = args_0
        elif isinstance(args_0, tuple):
            kwargs['wf'], kwargs['sr'] = args_0
    kwargs_keys = list(kwargs.keys())
    if 'filepath' in kwargs_keys:
        return wf_and_sr_from_filepath(filepath=kwargs['filepath'])
    if 'wf' in kwargs_keys:
        return kwargs['wf'], kwargs['sr']


def hear_sound(*args, **kwargs):
    wf, sr = wf_and_sr(*args, **kwargs)
    wf[random.randint(len(wf))] *= 1.001  # hack to avoid having exactly the same sound twice (creates an Audio bug!!)
    try:
        return Audio(data=wf, rate=sr, autoplay=kwargs.get('autoplay', False))
    except ValueError:
        try:
            # just return left audio (stereo PCM signals are unsupported
            return Audio(data=wf[0, :], rate=sr, autoplay=kwargs.get('autoplay', False))
        except:
            return Audio(data=wf[:, 0], rate=sr, autoplay=kwargs.get('autoplay', False))


def plot_wf(*args, **kwargs):
    wf, sr = wf_and_sr(*args, **kwargs)
    plt.plot(linspace(start=0, stop=len(wf) / float(sr), num=len(wf)), wf)


def display_sound(*args, **kwargs):
    plot_wf(*args, **kwargs)
    return hear_sound(*args, **kwargs)


def duration_of_wf_and_sr(wf, sr):
    return len(wf) / float(sr)


def n_wf_points_from_duration_and_sr(duration, sr):
    return int(round(duration * sr))


def get_consecutive_zeros_locations(wf, sr, thresh_consecutive_zeros_seconds=0.1):
    thresh_consecutive_zeros = thresh_consecutive_zeros_seconds * sr
    list_of_too_many_zeros_idx_and_len = list()
    cum_of_zeros = 0

    for i in range(len(wf)):
        if wf[i] == 0:
            cum_of_zeros += 1  # accumulate
        else:
            if cum_of_zeros > thresh_consecutive_zeros:
                list_of_too_many_zeros_idx_and_len.append({'idx': i - cum_of_zeros, 'len': cum_of_zeros})  # remember
            cum_of_zeros = 0  # reinit
    if cum_of_zeros > thresh_consecutive_zeros:
        list_of_too_many_zeros_idx_and_len.append({'idx': i - cum_of_zeros, 'len': cum_of_zeros})  # remember

    return list_of_too_many_zeros_idx_and_len


def crop_head_and_tail_silence(wf):
    assert len(wf.shape) == 1, "The silence crop is only implemented for mono sounds"
    first_non_zero = argmax(wf != 0)
    last_non_zero = len(wf) - argmin(flipud(wf == 0))
    return wf[first_non_zero:last_non_zero]


def is_wav_file(filepath):
    return os.path.splitext(filepath)[1] == '.wav'


def wav_file_framerate(file_pointer_or_path):
    if isinstance(file_pointer_or_path, str):
        file_pointer_or_path = wave.open(file_pointer_or_path)
        frame_rate = file_pointer_or_path.getframerate()
        file_pointer_or_path.close()
    else:
        frame_rate = file_pointer_or_path.getframerate()
    return frame_rate


def wf_and_sr_from_filepath(filepath, **kwargs):
    must_ensure_mono = kwargs.pop('ensure_mono', True)

    if is_wav_file(filepath):
        kwargs = dict({'always_2d': False}, **kwargs)
        if 'offset_s' in list(kwargs.keys()) or 'duration' in list(kwargs.keys()):
            sample_rate = wave.Wave_read(filepath).getframerate()
            start = int(round(kwargs.pop('offset_s', 0) * sample_rate))
            kwargs['start'] = start
            duration = kwargs.pop('duration', None)
            if duration is not None:
                kwargs['stop'] = int(start + round(duration * sample_rate))

        kwargs = filter_kwargs_to_func_arguments(sf.read, kwargs)
        wf, sr = sf.read(filepath, **kwargs)
    else:
        kwargs['offset'] = kwargs.pop('offset_s', 0.0)
        wf, sr = librosa.load(filepath, **kwargs)

    if must_ensure_mono:
        wf = ensure_mono(wf)
    return wf, sr

    # kwargs = dict({'sr': None}, **kwargs)
    # return librosa.load(filepath, **kwargs)


def wave_form(filepath, **kwargs):
    return wf_and_sr_from_filepath(filepath, **kwargs)[0]


def weighted_mean(yw1, yw2):
    common_len = min([len(yw1[0]), len(yw2[0])])
    a = yw1[0][:common_len]
    b = yw2[0][:common_len]
    return (a * yw1[1] + b * yw2[1]) / (yw1[1] + yw2[1])


def mk_transformed_copies_of_sound_files(source_path_iterator,
                                         file_reader=wf_and_sr_from_filepath,
                                         transform_fun=None,
                                         source_path_to_target_path=None,
                                         save_fun=None,
                                         onerror_fun=None):
    """
    Gets every filepath
        fed by source_path_iterator one by one,
        reads the file in with file_reader(filepath) to get a wave form and sample rate
        feeds these to the transform_fun(wf, sr), which returns another wf and sr,
        which are passed to save_fun(wf, sr, filepath) to be saved as a sound file,
        the target filepath being computed from source_path through the function source_path_to_target_path(path)
        If there's any errors and a onerror_fun(source_path, e) is given, it will be called instead of raising error
    """
    assert source_path_to_target_path is not None, "You must provide a save_fun (function or target folder)"
    if isinstance(source_path_to_target_path, str):
        target_folder = source_path_to_target_path
        assert os.path.exists(target_folder), \
            "The folder {} doesn't exist".format(target_folder)

        def source_path_to_target_path(source_path):
            source_name = os.path.splitext(os.path.basename(source_path))[0]
            return os.path.join(target_folder, source_name + '.wav')

    if save_fun is None:
        def save_fun(wf, sr, filepath):
            sf.write(file=filepath, data=wf, samplerate=sr)

    for source_path in source_path_iterator:
        try:
            wf, sr = file_reader(source_path)
            if transform_fun is not None:
                wf, sr = transform_fun(wf, sr)
            target_path = source_path_to_target_path(source_path)
            save_fun(wf=wf, sr=sr, filepath=target_path)
        except Exception as e:
            if onerror_fun is not None:
                onerror_fun(source_path, e)
            else:
                raise e


def plot_melspectrogram(spect_mat, sr=default_sr, hop_length=512, name=None):
    # Make a new figure
    plt.figure(figsize=(12, 4))
    # Display the spectrogram on a mel scale
    # sample rate and hop length parameters are used to render the time axis
    librosa.display.specshow(spect_mat, sr=sr, hop_length=hop_length, x_axis='time', y_axis='mel')
    # Put a descriptive title on the plot
    if name is not None:
        plt.title('mel power spectrogram of "{}"'.format(name))
    else:
        plt.title('mel power spectrogram')
    # draw a color bar
    plt.colorbar(format='%+02.0f dB')
    # Make the figure layout compact
    plt.tight_layout()


class Sound(object):
    def __init__(self, wf=None, sr=default_sr, name=''):
        if wf is None:
            wf = np.array([])
        self.wf = wf.copy()
        self.sr = sr
        self.name = name
        self.info = {}

    def copy(self):
        return Sound(wf=self.wf.copy(), sr=self.sr, name=self.name)

    ####################################################################################################################
    # CREATION

    @classmethod
    def from_file(cls, filepath, name=None, get_wav_info=False, **kwargs):
        """
        Construct sound object from sound file
        :param filepath: filepath of the sound file
        :param name: name to give this sound (will default to file name)
        :param kwargs: additional options, such as:
            * offset_s and duration (to retrieve only a segment of sound). Works with .wav file only
            * ensure_mono (if present and True (the default), will convert to mono)
        :return:
        """
        file_name, extension = os.path.splitext((os.path.basename(filepath)))
        name = name or file_name
        # kwargs = dict({'always_2d': False, 'ensure_mono': True}, **kwargs)

        wf, sr = wf_and_sr_from_filepath(filepath, **kwargs)

        if name is None:
            name = filepath

        sound = Sound(wf=wf, sr=sr, name=name)

        if get_wav_info and extension == '.wav':
            try:
                sound.info = sound_file_info_dict(filepath)
                offset_s = kwargs.get('offset_s', None)
                if offset_s is not None:
                    sound.info['offset_s'] = float(offset_s)
                duration = kwargs.get('duration', None)
                if duration is not None:
                    sound.info['duration'] = float(duration)
                if duration is not None or offset_s is not None:
                    offset_s = offset_s or 0
                    sound.info['frames'] = int((duration - offset_s) * default_sr)
                    sound.info.pop('size')
            except Exception:
                pass

        return sound

    @classmethod
    def from_(cls, sound):
        if isinstance(sound, tuple) and len(sound) == 2:  # then it's a (wf, sr) tuple
            return Sound(sound[0], sound[1])
        elif isinstance(sound, str) and os.path.isfile(sound):
            return Sound.from_file(sound)
        elif isinstance(sound, dict):
            if 'wf' in sound and 'sr' in sound:
                return Sound(sound['wf'], sound['sr'])
            else:
                return Sound.from_sref(sound)
        elif hasattr(sound, 'wf') and hasattr(sound, 'sr'):
            return Sound(sound.wf, sound.sr)
        else:
            try:
                return Sound.from_sound_iter(sound)
            except:
                raise TypeError("Couldn't figure out how that format represents sound")

    @classmethod
    def from_sref(cls, sref):
        wf, sr = wf_and_sr_from_filepath(**complete_sref(sref))
        # filepath = sref['filepath']
        # sample_rate = wave.Wave_read(sref['filepath']).getframerate()
        # start = int(round(sref.get('offset_s', 0) * sample_rate))
        # duration = sref.get('duration', None)
        # kwargs = {}
        # if duration is not None:
        #     kwargs = {'stop': int(start + round(duration * sample_rate))}
        # wf, sr = sf.read(filepath, always_2d=False, start=start, **kwargs)
        return Sound(wf, sr, name=sref.get('name', sref['filepath']))

    @classmethod
    def from_sound_iter(cls, sound_iter):
        wf = []
        for sound in sound_iter:
            if len(wf) == 0:
                sr = sound.sr
            wf.extend(list(sound.wf))

        return cls(wf=np.array(wf), sr=sr)

    @classmethod
    def from_sound_mix_spec(cls,
                            sound_mix_spec,
                            name='from_sound_mix_spec',
                            pre_normalization_function=lambda wf: wf / percentile(abs(wf), 95)):
        """
        Mix all sounds specified in the sound_mix_spec.

        A sound_mix_spec is an iterator that yields either of these formats:
            * a wave form
            * a Sound object
            * (This is the complete specification) a {sound, offset_s, weight} dict indicating
                offset_s (default 0 seconds): where the sound should be inserted
                weight (default 1): a weight, relative to the other sounds in the iterator, indicating whether the
                "volume" should be increased or decreased before mixing the sound

        Note: All wave forms are normalized before being multiplied by the given weight. The normalization function is
        given by the pre_normalization_function argument (default is no normalization)

        Note: If some of the sounds in the sound_mix_spec have different sample rates, they will be resampled to the
        sample rate of the first sound encountered. This process requires (not so fast) fast fourrier transform,
        so better have the same sample rate.
        """

        def _mk_sound_mix_spec(_sound_mix_spec):
            sound_mix_spec_default = dict(sound=None, offset_s=0, weight=1)
            _sound_mix_spec = _sound_mix_spec.copy()
            if isinstance(_sound_mix_spec, dict):  # if sound_mix_spec is a dict...
                if 'filepath' in list(_sound_mix_spec.keys()):  # ... and it has a 'filepath' key...
                    sref = _sound_mix_spec  # ... assume it's an sref...
                    sound = Sound.from_sref(sref)  # ... and get the sound from it, and make an actual sound_mix_spec
                    _sound_mix_spec = dict(sound_mix_spec_default, sound=sound)
                else:  # If it's not an sref...
                    sound = _sound_mix_spec['sound']  # ... assume it has a sound key
                    if isinstance(sound, dict):  # and if that "sound" is an sref, replace it by a actual sound object
                        _sound_mix_spec['sound'] = Sound.from_sref(_sound_mix_spec['sound'])
                        _sound_mix_spec = dict(sound_mix_spec_default, **_sound_mix_spec)
            elif isinstance(_sound_mix_spec, ndarray):
                _sound_mix_spec = dict(sound_mix_spec_default, sound=Sound(wf=_sound_mix_spec, sr=None))
            elif hasattr(_sound_mix_spec, 'wf'):
                _sound_mix_spec = dict(sound_mix_spec_default, sound=_sound_mix_spec)
            else:
                _sound_mix_spec = dict(sound_mix_spec_default, **_sound_mix_spec)
            _sound_mix_spec['sound'] = _sound_mix_spec[
                'sound'].copy()  # to make sure the we don't overwrite it in manip
            _sound_mix_spec['sound'].wf = ensure_mono(_sound_mix_spec['sound'].wf)
            # print(sound_mix_spec)
            return _sound_mix_spec

        # if the sound_iterator is a dict, take its values (ignore the keys)
        if isinstance(sound_mix_spec, dict):
            sound_mix_spec = list(sound_mix_spec.values())
        sound_mix_spec = iter(sound_mix_spec)
        # compute the weight factor. All input weights will be multiplied by this factor to avoid last sounds having
        # more volume than the previous ones

        # take the first sound as the sound to begin (and accumulate) with. As a result, the sr will be taken from there
        spec = _mk_sound_mix_spec(next(sound_mix_spec))
        result_sound = spec['sound']
        result_sound_sr = result_sound.sr  # will be the final sr, and all other sounds will be resampled to it
        result_sound.name = name
        result_sound.info = {}  # we don't want to keep the first sound's info around
        # offset the sound by required amount
        offset_length = ceil(spec.get('offset_s', 0) * result_sound_sr)
        result_sound.wf = prefix_with_silence(result_sound.wf, offset_length)

        # all subsequent weights should be multiplied by a weight_factor,
        # since the accumulating sound is considered to be of unit weight in the Sound.mix_in() method:
        weight_factor = 1 / spec.get('weight', 1.0)
        # initialize sound counter
        sounds_mixed_so_far = 1
        try:
            while True:
                spec = _mk_sound_mix_spec(next(sound_mix_spec))

                # resample sound to match self, if necessary
                #    (mix_in() method does it, but better do it before,
                #       because we're probably going to prefix this sound with silence, so it'll be longer)
                if spec['sound'].sr != result_sound.sr:
                    spec['sound'] = spec['sound'].resample(new_sr=result_sound.sr)

                # divide weight by number of sounds mixed so far, to avoid last sounds having more volume
                # than the previous ones
                weight = weight_factor * spec['weight'] / sounds_mixed_so_far
                # print(weight)
                # offset the new sound
                # print sound_mix_spec['sound_tag'], sound_mix_spec.get('offset_s')
                offset_length = ceil(spec.get('offset_s', 0) * result_sound_sr)
                spec['sound'].wf = prefix_with_silence(spec['sound'].wf, offset_length)

                # finally, mix these sounds
                result_sound.mix_in(spec['sound'],
                                    weight=weight,
                                    pre_normalization_function=pre_normalization_function)

                # increment the counter
                sounds_mixed_so_far += 1

        except StopIteration:
            pass

        return result_sound

    def save_to_wav(self, filepath=None, samplerate=None, **kwargs):
        samplerate = samplerate or self.sr
        filepath = filepath or (self.name + '.wav')
        sf.write(filepath, self.wf, samplerate=samplerate, **kwargs)

    ####################################################################################################################
    # TRANSFORMATIONS

    def ensure_mono(self):
        self.wf = ensure_mono(self.wf)

    def resample(self, new_sr, inplace=False):
        new_wf = resample_wf(self.wf, self.sr, new_sr)
        if not inplace:
            return Sound(wf=new_wf, sr=new_sr, name=self.name)
        else:
            self.wf = new_wf
            self.sr = new_sr

    def crop_with_idx(self, first_idx, last_idx):
        """
        Crop with frame indices.
        :param first_idx: First frame index (starting with 0, like with lists)
        :param last_idx: Last frame index. Like with list indices again. If frame n is actually the (n+1)th frame...
        :return:
        """
        cropped_sound = self.copy()
        cropped_sound.wf = cropped_sound.wf[first_idx:last_idx]
        return cropped_sound

    # def __getitem__

    def crop_with_seconds(self, first_second, last_second):
        return self.crop_with_idx(int(round(first_second * self.sr)), int(round(last_second * self.sr)))

    def mix_in(self,
               sound,
               weight=1,
               pre_normalization_function=lambda wf: wf / percentile(abs(wf), 95)):

        # resample sound to match self, if necessary
        if sound.sr != self.sr:
            sound = sound.resample(new_sr=self.sr)
        new_wf = sound.wf.copy()

        # suffix the shortest sound with silence to match lengths
        existing_sound_length = len(self.wf)
        new_sound_length = len(new_wf)
        length_difference = existing_sound_length - new_sound_length
        if length_difference > 0:
            new_wf = suffix_with_silence(new_wf, length_difference)
        elif length_difference < 0:
            self.wf = suffix_with_silence(self.wf, -length_difference)

        # mix the new wf into self.wf
        # print(pre_normalization_function(arange(100)))
        self.wf = weighted_mean([pre_normalization_function(self.wf), 1],
                                [pre_normalization_function(new_wf), weight])

    def append(self, sound, glue=0.0):
        assert sound.sr == self.sr, "you can only append sounds if they have the same sample rate at this point"
        if isinstance(glue, (float, int)):
            n_samples = int(glue * self.sr)
            glue = zeros(n_samples)
        self.wf = hstack((self.wf, glue, sound.wf))

    def melspectr_matrix(self, **mel_kwargs):
        mel_kwargs = dict({'n_fft': 2048, 'hop_length': 512, 'n_mels': 128}, **mel_kwargs)
        S = librosa.feature.melspectrogram(np.array(self.wf).astype(float), sr=self.sr, **mel_kwargs)
        # Convert to log scale (dB). We'll use the peak power as reference.
        return librosa.amplitude_to_db(S, ref=np.max)

    ####################################################################################################################
    # DISPLAY FUNCTIONS

    def plot_wf(self, **kwargs):
        kwargs = dict(alpha=0.8, **kwargs)
        plot_wf(wf=self.wf.copy(), sr=self.sr, **kwargs)

    def hear(self, autoplay=False, **kwargs):
        wf = np.array(ensure_mono(self.wf)).astype(float)
        wf[np.random.randint(
            len(wf))] *= 1.001  # hack to avoid having exactly the same sound twice (creates an Audio bug)
        return Audio(data=wf, rate=self.sr, autoplay=autoplay, **kwargs)

    def display_sound(self, **kwargs):
        self.plot_wf()
        return self.hear(**kwargs)

    def display(self, sound_plot='mel', autoplay=False, **kwargs):
        """

        :param sound_plot: 'mel' (default) to plot melspectrogram, 'wf' to plot wave form, and None to plot nothing at all
        :param kwargs:
        :return:
        """
        if sound_plot == 'mel':
            self.melspectrogram(plot_it=True, **kwargs)
        elif sound_plot == 'wf':
            self.plot_wf(**kwargs)
        return self.hear(autoplay=autoplay)

    def melspectrogram(self, plot_it=False, **mel_kwargs):
        mel_kwargs = dict({'n_fft': 2048, 'hop_length': 512, 'n_mels': 128}, **mel_kwargs)
        log_S = self.melspectr_matrix(**mel_kwargs)
        if plot_it:
            plot_melspectrogram(log_S, sr=self.sr, hop_length=mel_kwargs['hop_length'])
        return log_S

    ####################################################################################################################
    # MISC
    def duration(self):
        return duration_of_wf_and_sr(self.wf, self.sr)

    def wf_sr_dict(self):
        return {'wf': self.wf, 'sr': self.sr}

    def wf_sr_tuple(self):
        return self.wf, self.sr



__author__ = 'thor'

from numpy import *
import numpy as np
import librosa
from sklearn.base import BaseEstimator, TransformerMixin

from snippyng.util import resample_wf
from snippyng.util import Sound
import matplotlib.pyplot as plt

from sklearn.decomposition import IncrementalPCA
from sklearn.preprocessing import MinMaxScaler, StandardScaler


class DeAmplitudedEigenSound(BaseEstimator, TransformerMixin):
    """
    Transforms sound waves to eigen decomposition of the log of the melspectrogram minus the log of the amplitude sums
    (the columns of the melspectrogram) enhanced with the differential of the log of the amplitude array.

    The idea is to normalize the melspectrogram columns by the amplitude, so to reduce non-meaningful spectrogram
    variation, but keeping the sequence of amplitude around (because in itself, it carries a lot of meaning).

    An (incremental) decomposition of the matrix formed from the normalized log10 melspectrogram will be fit,
    and the fit will also learn the mean and variance of the diff of the amplitude sequence (which also carries meaning)
    """
    def __init__(self, spectr_decomp=IncrementalPCA(),
                 amp_diff_scaler=StandardScaler(),
                 mel_kwargs={'n_fft': 2048, 'hop_length': 512, 'n_mels': 128},
                 common_sr=44100,
                 min_amp=1e-10):
        self.spectr_decomp = spectr_decomp
        self.amp_diff_scaler = amp_diff_scaler
        self.mel_kwargs = mel_kwargs
        self.common_sr = common_sr
        self.min_amp = min_amp
        self.log10_min_amp = log10(min_amp)

    def melspectrogram(self, wf, sr):
        if sr != self.common_sr:
            wf = resample_wf(wf, sr, new_sr=self.common_sr)
        return librosa.feature.melspectrogram(wf, sr=self.common_sr, **self.mel_kwargs)

    def deamplituded_mel(self, mel_spectrogram):
        mel_spectrogram = np.maximum(mel_spectrogram, self.min_amp)
        amplitude_seq = log10(np.sum(mel_spectrogram, axis=0))
        return log10(mel_spectrogram) - amplitude_seq, diff(amplitude_seq)

    def partial_fit(self, X, y=None):
        _deamplituded_mel, diff_amplitude_seq = self.deamplituded_mel(X)
        self.spectr_decomp.partial_fit(_deamplituded_mel.T)
        self.amp_diff_scaler.partial_fit(diff_amplitude_seq.reshape(-1, 1))
        return self

    def wf_partial_fit(self, wf, sr):
        return self.partial_fit(self.melspectrogram(wf, sr))

    def transform(self, X):
        _deamplituded_mel, diff_amplitude_seq = self.deamplituded_mel(X)
        return np.hstack(
            (
                self.spectr_decomp.transform(_deamplituded_mel.T),
                self.amp_diff_scaler.transform(np.vstack((self.log10_min_amp, diff_amplitude_seq.reshape(-1, 1))))
            )
        )

    def inverse_transform(self, X):
        eigen_mat = X[:, :-1]
        mel_spectrogram = self.spectr_decomp.inverse_transform(eigen_mat)

        amp_diff = self.amp_diff_scaler.inverse_transform(X[:, -1])
        amplitude_seq = cumsum(amp_diff)

        return (mel_spectrogram.T + amplitude_seq.T).T


    def transform_as_dict(self, X):
        _deamplituded_mel, diff_amplitude_seq = self.deamplituded_mel(X)
        return {
            'scaled_amp_diff': self.amp_diff_scaler.transform(
                np.vstack((self.log10_min_amp, diff_amplitude_seq.reshape(-1, 1)))),
            'eigen': self.spectr_decomp.transform(_deamplituded_mel.T)
        }

    def wf_transform(self, wf, sr):
        return self.transform(self.melspectrogram(wf, sr))

    def wf_transform_as_dict(self, wf, sr):
        return self.transform_as_dict(self.melspectrogram(wf, sr))

    def plot_transform_for_wf(self, wf, sr, **kwargs):
        kwargs.setdefault('aspect', 'auto')
        kwargs.setdefault('origin', 'lower')
        kwargs.setdefault('interpolation', 'nearest')

        fp = self.wf_transform_as_dict(wf, sr)
        eigens = fp['eigen'].T
        kwargs.setdefault('cmap', plt.get_cmap('coolwarm'))

        fig, ax1 = plt.subplots()

        plt.imshow(eigens, axes=ax1, **kwargs)
        plt.yticks([])

        ax2 = ax1.twinx()
        ax2.plot(cumsum(fp['scaled_amp_diff']), 'k-')
        plt.yticks([])
        plt.axis('tight')



class EigenSound(object):
    def __init__(self, pca_components, mel_kwargs, n_components=None):
        if n_components is not None:
            self.n_components = n_components
            self.pca_components = pca_components[:n_components, :]
        else:
            self.n_components = shape(pca_components)[0]
            self.pca_components = pca_components
        self.mel_kwargs = mel_kwargs

    def melspectr(self, sound):
        if isinstance(sound, Sound):
            sound = (sound.wf, sound.sr)
        if isinstance(sound, tuple):
            # then melspectr is actually the original sound waveform (wf, sr), so get the melspectr
            sound = librosa.feature.melspectrogram(y=sound[0], sr=sound[1], **self.mel_kwargs)
        return sound

    def transform_melspectr(self, sound, display=False):
        eigen_spectr = dot(self.pca_components, log(self.melspectr(sound)))
        if display == True:
            h = plt.matshow(np.flipud(eigen_spectr), cmap='hot_r')
            # h = librosa.display.specshow(data=eigen_spectr, y_axis='mel')
            # plt.colorbar(format='%+02.01f dB')
            return h
        else:
            return eigen_spectr

    def inverse_transform_melspectr(self, eigen_spectr, display=False):
        melspectr = dot(self.pca_components.T, eigen_spectr)
        if display == True:
            h = librosa.display.specshow(data=melspectr, y_axis='mel')
            plt.colorbar(format='%+02.01f dB')
            return h
        else:
            return melspectr

    def plot_two_first_eigens(self, sound, title=True, plot_kwargs={}):
        X = self.transform_melspectr(sound);
        plot_kwargs = dict(plot_kwargs, **{'alpha': 0.3})
        plt.plot(X[0, :], X[1, :], '-o', **plot_kwargs)
        plt.annotate('begining', xy=(X[0, 0], X[1, 0]))
        plt.annotate('end', xy=(X[0, -1], X[1, -1]))
        if title:
            if title is True:
                title = "two first eigensounds"
                if isinstance(sound, Sound):
                    if sound.name:
                        title += "\n of {}".format(sound.name)
            plt.title(title)


class GeneralEigenSound(object):
    def __init__(self, decomp, mel_kwargs):
        self.decomp = decomp
        self.mel_kwargs = mel_kwargs

    def melspectr(self, sound):
        if isinstance(sound, Sound):
            sound = (sound.wf, sound.sr)
        if isinstance(sound, tuple):
            # then melspectr is actually the original sound waveform (wf, sr), so get the melspectr
            sound = librosa.feature.melspectrogram(y=sound[0], sr=sound[1], **self.mel_kwargs)
        return sound

    def transform_melspectr(self, sound, display=False):
        # eigen_spectr = dot(self.decomp.components_, log(self.melspectr(sound)))
        eigen_spectr = self.decomp.transform(log(self.melspectr(sound)).T).T
        if display == True:
            h = plt.matshow(np.flipud(eigen_spectr), cmap='hot_r')
            return h
        else:
            return eigen_spectr

    def inverse_transform_melspectr(self, eigen_spectr, display=False):
        melspectr = dot(self.pca_components.T, eigen_spectr)
        if display == True:
            h = librosa.display.specshow(data=melspectr, y_axis='mel')
            plt.colorbar(format='%+02.01f dB')
            return h
        else:
            return melspectr

    def plot_two_first_eigens(self, sound, title=True, plot_kwargs={}):
        X = self.transform_melspectr(sound);
        plot_kwargs = dict(plot_kwargs, **{'alpha': 0.3})
        plt.plot(X[0, :], X[1, :], '-o', **plot_kwargs)
        plt.annotate('begining', xy=(X[0, 0], X[1, 0]))
        plt.annotate('end', xy=(X[0, -1], X[1, -1]))
        if title:
            if title is True:
                title = "two first eigensounds"
                if isinstance(sound, Sound):
                    if sound.name:
                        title += "\n of {}".format(sound.name)
            plt.title(title)